chaosPower = CreateFrame("Frame", "chaosPower", UIParent)
chaosPower:SetScript("OnEvent", function(self, event, ...) self[event](self, event, ...) end)

----------------------------------------------------------------------------------------
local chaosPowerDisplay = nil
local chaosBoltPowerPrediction = 0
local unlocked = false
local inCombat = false
local timeLeftMaxWidth = 0

local defaults =
{
  posX = 50,
  posY = 300,
  hideOutOfCombat = true,
  width = 60,
  height = 20,
  fontSize = 12,
  showIcon = true,
  showTimeLeft = true,
  timeLeftHeight = 3,
}

local trackedAuras =
{
  ----------------------------------------------------
  -- Warlock Spells/Buffs/Proccs
  113858, -- Dark Soul: Instability

  ----------------------------------------------------
  -- Weapon Enchants
  104993, -- Jade Spirit
  104510, -- Windsong (Mastery)
  104509, -- Windsong (Crit)

  ----------------------------------------------------
  -- Sets
  145164, -- Ember Master (T16 4P)

  ----------------------------------------------------
  -- Trinket Proccs (T15)
  138788, -- Wushoolay's Final Choice
  138898, -- Breath of the Hydra
  139133, -- Cha-Ye's Essence of Brilliance
  138963, -- Unerring Vision of Lei Shen

  ----------------------------------------------------
  -- Trinket Proccs (T16)
  146046, -- Purified Bindings of Immerseus
  148906, -- Kardris' Toxic Totem
  148897, -- Frenzied Crystal of Rage
  146202, -- Black Blood of Y'Shaarj
}

chaosPowerDB = {}

----------------------------------------------------------------------------------------
chaosPower:RegisterEvent("ADDON_LOADED")
chaosPower:RegisterEvent("PLAYER_LOGIN")
chaosPower:RegisterEvent("PLAYER_REGEN_ENABLED")
chaosPower:RegisterEvent("PLAYER_REGEN_DISABLED")

chaosPower:RegisterEvent("COMBAT_RATING_UPDATE")
chaosPower:RegisterEvent("SPELL_POWER_CHANGED")
chaosPower:RegisterEvent("UNIT_STATS")
chaosPower:RegisterEvent("PLAYER_DAMAGE_DONE_MODS")

----------------------------------------------------------------------------------------
function chaosPower:COMBAT_RATING_UPDATE()    self:UpdateChaosBoltPowerPrediction() end
function chaosPower:SPELL_POWER_CHANGED()     self:UpdateChaosBoltPowerPrediction() end
function chaosPower:UNIT_STATS()              self:UpdateChaosBoltPowerPrediction() end
function chaosPower:PLAYER_DAMAGE_DONE_MODS() self:UpdateChaosBoltPowerPrediction() end

----------------------------------------------------------------------------------------
local function LoadDefaults()
  chaosPowerDB = chaosPowerDB or {}
  for k,v in pairs(defaults) do
    if(type(chaosPowerDB[k]) == 'nil') then chaosPowerDB[k] = v end
  end
end

----------------------------------------------------------------------------------------
function chaosPower:ADDON_LOADED(event, addon)

  if (addon ~= "ChaosPower") then return end
  self:UnregisterEvent(event)
  LoadDefaults()

  chaosPowerDisplay = CreateFrame("Frame", "chaosPowerDisplay", UIParent)
  chaosPowerDisplay:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", chaosPowerDB.posX, chaosPowerDB.posY)
  chaosPowerDisplay:SetMovable(true)
  chaosPowerDisplay:EnableMouse(false)
  chaosPowerDisplay:RegisterForDrag("LeftButton")

  local cpt = chaosPowerDisplay:CreateFontString(nil, "OVERLAY")
  cpt:SetJustifyH("CENTER")
  chaosPowerDisplay.text = cpt

  local tex = chaosPowerDisplay:CreateTexture(nil, "BACKGROUND")
  tex:SetAllPoints(chaosPowerDisplay)
  chaosPowerDisplay.texture = tex

  local icon = chaosPowerDisplay:CreateTexture(nil, "LOW")
  icon:SetTexture(select(3, GetSpellInfo(116858)))
  icon:SetPoint("BOTTOMLEFT")
  chaosPowerDisplay.icon = icon

  local timeLeftBar = chaosPowerDisplay:CreateTexture(nil, "LOW")
  timeLeftBar:SetDrawLayer("BACKGROUND", 3)
  chaosPowerDisplay.timeLeftBar = timeLeftBar
  chaosPowerDisplay:SetScript("OnUpdate", chaosPower.UpdateTimeLeftOnCurrentStrength)

  local timeLeftBar2 = chaosPowerDisplay:CreateTexture(nil, "LOW")
  timeLeftBar2:SetDrawLayer("BACKGROUND", 2)
  chaosPowerDisplay.timeLeftBar2 = timeLeftBar2

  local cbcIndicator = chaosPowerDisplay:CreateTexture(nil, "LOW")
  cbcIndicator:SetDrawLayer("BACKGROUND", 4)
  cbcIndicator:SetWidth(2)
  chaosPowerDisplay.cbCastTimeIndicator = cbcIndicator

  self:UpdateDisplayFrame()

  _, playerClass = UnitClass("player")
  if (playerClass ~= "WARLOCK") then
    self:UnregisterAllEvents()
    chaosPowerDisplay:Hide()
  end
end

----------------------------------------------------------------------------------------
function chaosPower:PLAYER_LOGIN()
  self:UpdateChaosBoltPowerPrediction()
end

----------------------------------------------------------------------------------------
function chaosPower:PLAYER_REGEN_ENABLED()
  inCombat = false;
  if chaosPowerDB.hideOutOfCombat then chaosPowerDisplay:Hide() end
end

----------------------------------------------------------------------------------------
function chaosPower:PLAYER_REGEN_DISABLED()
  inCombat = true;
  if chaosPowerDB.hideOutOfCombat then chaosPowerDisplay:Show() end
end

----------------------------------------------------------------------------------------
function chaosPower:UpdateChaosBoltPowerPrediction()
  local mastery, crit, spelldmg = GetMastery(), GetSpellCritChance(6), GetSpellBonusDamage(6)
  if crit > 100 then crit = 100 end

  chaosBoltPowerPrediction = floor((2400 + 2.25 * spelldmg) * (1 + crit / 100) * (1 + mastery / 100) / 1000 + .5)
  chaosPowerDisplay.text:SetText(chaosBoltPowerPrediction)
end

----------------------------------------------------------------------------------------
function chaosPower:AuraIsTracked(spellId)
  if (spellId == nil) then return false end

  for _,id in pairs(trackedAuras) do
    if id == spellId then return true end
  end
  return false
end

----------------------------------------------------------------------------------------
function chaosPower:UpdateTimeLeftOnCurrentStrength()
  if (not chaosPowerDB.showTimeLeft) then return end
  local shortestTimeRemaining = -1
  local shortestTimeRemaining2 = -1

  for i = 1, 40 do
    local _,_,_,_,_,_,remainingTime,_,_,_,spellId = UnitAura("player", i, "HELPFUL")
    if (remainingTime ~= nil and chaosPower:AuraIsTracked(spellId)) then
      remainingTime = remainingTime - GetTime()
      if (remainingTime > 0) then
        if (remainingTime < shortestTimeRemaining or shortestTimeRemaining == -1) then
          shortestTimeRemaining2 = shortestTimeRemaining
          shortestTimeRemaining = remainingTime
        elseif (remainingTime < shortestTimeRemaining2 or shortestTimeRemaining2 == -1) then
          shortestTimeRemaining2 = remainingTime
        end
      end
    end
  end

  if (shortestTimeRemaining == -1) then
    chaosPowerDisplay.timeLeftBar:Hide()
    chaosPowerDisplay.cbCastTimeIndicator:Hide()
  else
    local barWidth = math.min(shortestTimeRemaining / 12 * timeLeftMaxWidth, timeLeftMaxWidth)
    chaosPowerDisplay.timeLeftBar:SetWidth(barWidth)
    chaosPowerDisplay.timeLeftBar:Show()

    local cbCastTime = select(7, GetSpellInfo(116858)) / 1000
    local _,_,_,_,_,remainingCastTime = UnitCastingInfo("player")

    local cbcIndicatorPosX = 0
    if (remainingCastTime ~= nil) then
      remainingCastTime = remainingCastTime / 1000 - GetTime()
      cbcIndicatorPosX = (remainingCastTime + cbCastTime) / 12 * timeLeftMaxWidth
    else
      cbcIndicatorPosX = cbCastTime / 12 * timeLeftMaxWidth
    end
    chaosPowerDisplay.cbCastTimeIndicator:SetPoint("BOTTOMLEFT", chaosPowerDisplay.timeLeftBar, "BOTTOMLEFT", cbcIndicatorPosX - 1, 0)
    chaosPowerDisplay.cbCastTimeIndicator:Show()

    if (cbcIndicatorPosX >= barWidth) then
      chaosPowerDisplay.timeLeftBar:SetTexture(.7, 0, 0, .9)
    else
      chaosPowerDisplay.timeLeftBar:SetTexture(0, .5, 0, .9)
    end
  end

  if (shortestTimeRemaining2 == -1) then
    chaosPowerDisplay.timeLeftBar2:Hide()
  else
    local barWidth = math.min(shortestTimeRemaining2 / 12 * timeLeftMaxWidth, timeLeftMaxWidth)
    chaosPowerDisplay.timeLeftBar2:SetWidth(barWidth)
    chaosPowerDisplay.timeLeftBar2:Show()
    chaosPowerDisplay.timeLeftBar2:SetTexture(.5, .5, .5, .45)
  end
end

----------------------------------------------------------------------------------------
function chaosPower:UpdateDisplayFrame()
  chaosPowerDisplay:SetWidth(chaosPowerDB.width)
  chaosPowerDisplay:SetHeight(chaosPowerDB.height)

  ---------------------
  -- background
  chaosPowerDisplay.texture:SetTexture(0, 0, 0, .7)

  ---------------------
  -- text
  chaosPowerDisplay.text:SetFont("Fonts\\FRIZQT__.TTF", chaosPowerDB.fontSize)
  chaosPowerDisplay.text:SetShadowColor(0, 0, 0)
  chaosPowerDisplay.text:SetShadowOffset(0.8, -0.8)
  chaosPowerDisplay.text:SetTextColor(1, 1, 1)
  chaosPowerDisplay.text:SetText(chaosBoltPowerPrediction)

  chaosPowerDisplay.text:ClearAllPoints()
  local textPosOffsetY = chaosPowerDB.showTimeLeft and chaosPowerDB.timeLeftHeight / 2 or 0
  if (chaosPowerDB.showIcon) then
    chaosPowerDisplay.text:SetPoint("RIGHT", -3, textPosOffsetY)
  else
    chaosPowerDisplay.text:SetPoint("CENTER", 0, textPosOffsetY)
  end

  ---------------------
  -- icon
  chaosPowerDisplay.icon:SetWidth(chaosPowerDB.height)
  chaosPowerDisplay.icon:SetHeight(chaosPowerDB.height)
  if (chaosPowerDB.showIcon) then
    chaosPowerDisplay.icon:Show()
  else
    chaosPowerDisplay.icon:Hide()
  end

  ---------------------
  -- time left bar
  chaosPowerDisplay.timeLeftBar:SetHeight(chaosPowerDB.timeLeftHeight)
  chaosPowerDisplay.timeLeftBar2:SetHeight(chaosPowerDB.timeLeftHeight)
  if (chaosPowerDB.showIcon) then
    chaosPowerDisplay.timeLeftBar:SetPoint("BOTTOMLEFT", chaosPowerDisplay.icon, "BOTTOMRIGHT")
    chaosPowerDisplay.timeLeftBar2:SetPoint("BOTTOMLEFT", chaosPowerDisplay.icon, "BOTTOMRIGHT")
    timeLeftMaxWidth = chaosPowerDB.width - chaosPowerDB.height
  else
    chaosPowerDisplay.timeLeftBar:SetPoint("BOTTOMLEFT")
    chaosPowerDisplay.timeLeftBar2:SetPoint("BOTTOMLEFT")
    timeLeftMaxWidth = chaosPowerDB.width
  end

  if (chaosPowerDB.showTimeLeft) then
    chaosPowerDisplay.timeLeftBar:Show()
    chaosPowerDisplay.timeLeftBar2:Show()
  else
    chaosPowerDisplay.timeLeftBar:Hide()
    chaosPowerDisplay.timeLeftBar2:Hide()
  end

  ---------------------
  -- chaos bolt cast time indicator
  chaosPowerDisplay.cbCastTimeIndicator:SetTexture(1, 1, 1, 1)
  chaosPowerDisplay.cbCastTimeIndicator:SetHeight(chaosPowerDB.timeLeftHeight + 1)

  ---------------------
  -- hide out of combat
  if chaosPowerDB.hideOutOfCombat and not inCombat then
    chaosPowerDisplay:Hide()
  end
end

----------------------------------------------------------------------------------------
local function SetFrameMovable(f, v)
  f:EnableMouse(v)
  f:SetScript("OnDragStart", v and f.StartMoving or nil)
  f:SetScript("OnDragStop", v and
    function(self)
      self:StopMovingOrSizing()
      chaosPowerDB.posX = self:GetLeft()
      chaosPowerDB.posY = self:GetBottom()
    end
  or nil)
  f:SetScript("OnHide", v and f.StopMovingOrSizing or nil)
end

----------------------------------------------------------------------------------------
local function StatusMsg(str1, str2, data, name, short)
  local R,G,t = '|cFFFF0000', '|cFF00FF00', ''
  if data ~= nil then t = data and G..(short and 'on|r' or 'enabled|r') or R..(short and 'off|r' or 'disabled|r') end
  t = (name and '|cFFFFFF00ChaosPower:|r ' or '')..str1..t..str2
  ChatFrame1:AddMessage(t)
end

----------------------------------------------------------------------------------------
local function StatusMsgVal(str1, str2, data, name)
  local G,t = '|cFF00FF00', ''
  if (data ~= nil) then t = G..data..'|r' end
  t = (name and '|cFFFFFF00ChaosPower:|r ' or '')..str1..t..str2
  ChatFrame1:AddMessage(t)
end

----------------------------------------------------------------------------------------
local function HandleSlash(str)
  local str, str2 = strsplit(" ", str)
  if str == 'move' then
    unlocked = not unlocked
    SetFrameMovable(chaosPowerDisplay, unlocked)
    if not inCombat and chaosPowerDB.hideOutOfCombat then
      if unlocked then
        chaosPowerDisplay:Show()
      else
        chaosPowerDisplay:Hide()
      end
    end
    StatusMsg('Frame lock is now ', '.', not unlocked, true, false)

  elseif str == 'hideooc' then
    chaosPowerDB.hideOutOfCombat = not chaosPowerDB.hideOutOfCombat
    if not inCombat then
      if chaosPowerDB.hideOutOfCombat then
        chaosPowerDisplay:Hide()
      else
        chaosPowerDisplay:Show()
      end
    end
    StatusMsg('Hiding the frame out of combat is now ', '.', chaosPowerDB.hideOutOfCombat, true, false)

  elseif str == 'width' then
    value = tonumber(str2)
    if value then
      chaosPowerDB.width = value
      chaosPower:UpdateDisplayFrame()
      StatusMsgVal('Width has been set to ', '.', chaosPowerDB.width, true)
    else
      StatusMsg('You have to specify a value, e.g. /cp width 60.', '', nil, true, false)
    end

  elseif str == 'height' then
    value = tonumber(str2)
    if value then
      chaosPowerDB.height = value
      chaosPower:UpdateDisplayFrame()
      StatusMsgVal('Width has been set to ', '.', chaosPowerDB.height, true)
    else
      StatusMsg('You have to specify a value, e.g. /cp height 20.', '', nil, true, false)
    end

  elseif str == 'fontsize' then
    value = tonumber(str2)
    if value then
      chaosPowerDB.fontSize = value
      chaosPower:UpdateDisplayFrame()
      StatusMsgVal('Width has been set to ', '.', chaosPowerDB.fontSize, true)
    else
      StatusMsg('You have to specify a value, e.g. /cp fontsize 12.', '', nil, true, false)
    end

  elseif str == 'icon' then
    chaosPowerDB.showIcon = not chaosPowerDB.showIcon
    chaosPower:UpdateDisplayFrame()
    StatusMsg('The spell icon is now ', '.', chaosPowerDB.showIcon, true, false)

  elseif str == 'timeleft' then
    chaosPowerDB.showTimeLeft = not chaosPowerDB.showTimeLeft
    chaosPower:UpdateDisplayFrame()
    StatusMsg('The time left bar is now ', '.', chaosPowerDB.showTimeLeft, true, false)

  elseif str == 'timeleftheight' then
    chaosPowerDB.timeLeftHeight = tonumber(str2)
    chaosPower:UpdateDisplayFrame()
    StatusMsgVal('The height of the time left bar has been set to ', '.', chaosPowerDB.timeLeftHeight, true)

  else
    ChatFrame1:AddMessage('|cFFFFFF00ChaosPower:|r')
    StatusMsg('(', ') |cFFFFFF00move|r - Toggle movable status.', unlocked, false, true)
    StatusMsg('(', ') |cFFFFFF00hideooc|r - Toggle hiding the frame out of combat.', chaosPowerDB.hideOutOfCombat, false, true)
    StatusMsgVal('(', ') |cFFFFFF00width|r - Set the width of the frame.', chaosPowerDB.width, false)
    StatusMsgVal('(', ') |cFFFFFF00height|r - Set the height of the frame.', chaosPowerDB.height, false)
    StatusMsgVal('(', ') |cFFFFFF00fontsize|r - Set the font size.', chaosPowerDB.fontSize, false)
    StatusMsg('(', ') |cFFFFFF00icon|r - Show or hide the spell icon.', chaosPowerDB.showIcon, false, true)
    StatusMsg('(', ') |cFFFFFF00timeleft|r - Toggle a bar displaying the time left on current strength.', chaosPowerDB.showTimeLeft, false, true)
    StatusMsgVal('(', ') |cFFFFFF00timeleftheight|r - Set the height of the time left bar.', chaosPowerDB.timeLeftHeight, false)
  end
end

----------------------------------------------------------------------------------------
SLASH_CHAOSPOWER1 = '/chaosPower'
SLASH_CHAOSPOWER2 = '/cp'
SlashCmdList.CHAOSPOWER = HandleSlash